@@ -10,6 +10,7 @@
 # distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
 # ANY KIND, either express or implied. See the License for the specific
 # language governing permissions and limitations under the License.
+import base64
 import re
 
 from tests import temporary_file
@@ -18,7 +19,7 @@ from nose.tools import assert_equal
 
 import botocore.session
 from botocore.config import Config
-from botocore.compat import datetime, urlsplit, parse_qs
+from botocore.compat import datetime, urlsplit, parse_qs, get_md5
 from botocore.exceptions import ParamValidationError, ClientError
 from botocore.exceptions import InvalidS3UsEast1RegionalEndpointConfigError
 from botocore.parsers import ResponseParserError
@@ -562,12 +563,24 @@ class TestAccesspointArn(BaseS3ClientConfigurationTest):
         accesspoint_arn = (
             'arn:aws:s3:us-west-2:123456789012:accesspoint:myendpoint'
         )
-        self.client, _ = self.create_stubbed_s3_client(
+        self.client, http_stubber = self.create_stubbed_s3_client(
             endpoint_url='https://custom.com')
-        with self.assertRaises(
-                botocore.exceptions.
-                UnsupportedS3AccesspointConfigurationError):
-            self.client.list_objects(Bucket=accesspoint_arn)
+        http_stubber.add_response()
+        self.client.list_objects(Bucket=accesspoint_arn)
+        expected_endpoint = 'myendpoint-123456789012.custom.com'
+        self.assert_endpoint(http_stubber.requests[0], expected_endpoint)
+
+    def test_accesspoint_arn_with_custom_endpoint_and_dualstack(self):
+        accesspoint_arn = (
+            'arn:aws:s3:us-west-2:123456789012:accesspoint:myendpoint'
+        )
+        self.client, http_stubber = self.create_stubbed_s3_client(
+            endpoint_url='https://custom.com',
+            config=Config(s3={'use_dualstack_endpoint': True}))
+        http_stubber.add_response()
+        self.client.list_objects(Bucket=accesspoint_arn)
+        expected_endpoint = 'myendpoint-123456789012.custom.com'
+        self.assert_endpoint(http_stubber.requests[0], expected_endpoint)
 
     def test_accesspoint_arn_with_s3_accelerate(self):
         accesspoint_arn = (
@@ -745,6 +758,24 @@ class TestAccesspointArn(BaseS3ClientConfigurationTest):
         )
         self.assert_endpoint(request, expected_endpoint)
 
+    def test_basic_outpost_arn(self):
+        outpost_arn = (
+            'arn:aws:s3-outposts:us-west-2:123456789012:outpost:'
+            'op-01234567890123456:accesspoint:myaccesspoint'
+        )
+        self.client, self.http_stubber = self.create_stubbed_s3_client(
+            endpoint_url='https://custom.com',
+            region_name='us-east-1')
+        self.http_stubber.add_response()
+        self.client.list_objects(Bucket=outpost_arn)
+        request = self.http_stubber.requests[0]
+        self.assert_signing_name(request, 's3-outposts')
+        self.assert_signing_region(request, 'us-west-2')
+        expected_endpoint = (
+            'myaccesspoint-123456789012.op-01234567890123456.custom.com'
+        )
+        self.assert_endpoint(request, expected_endpoint)
+
     def test_outpost_arn_with_s3_accelerate(self):
         outpost_arn = (
             'arn:aws:s3-outposts:us-west-2:123456789012:outpost:'
@@ -2216,3 +2247,21 @@ def _verify_presigned_url_addressing(region, bucket, key, s3_config,
     parts = urlsplit(url)
     actual = '%s://%s%s' % parts[:3]
     assert_equal(actual, expected_url)
+
+
+class TestS3DeleteObjects(BaseS3OperationTest):
+    def test_escape_keys_in_xml_payload(self):
+        self.http_stubber.add_response()
+        with self.http_stubber:
+            response = self.client.delete_objects(
+                Bucket='mybucket',
+                Delete={
+                    'Objects': [{'Key': 'some\r\n\rkey'}]
+                },
+            )
+        request = self.http_stubber.requests[0]
+        self.assertNotIn(b'\r\n\r', request.body)
+        self.assertIn(b'&#xD;&#xA;&#xD;', request.body)
+        content_md5_bytes = get_md5(request.body).digest()
+        content_md5 = base64.b64encode(content_md5_bytes)
+        self.assertEqual(content_md5, request.headers['Content-MD5'])
